@@ -32,6 +32,7 @@
 #include "BTree.h"
 #include "BTreePrivate.h"
 #include "CheckHFS.h"
+#include "BTreeScanner.h"
 
 #ifdef __cplusplus
 extern	"C" {
@@ -49,6 +50,11 @@ enum {
 	kNoHint						= 0
 };
 
+
+enum {
+	kHFSRepairCatalogFileID = (kHFSBogusExtentFileID - 1) /* used to repair catalog file */
+};
+
 //
 // Misc constants
 //
@@ -66,14 +72,16 @@ enum {
 #define	VBM_FNum	2				/* file number representing the volume bit map */
 #define	MDB_BlkN	2				/* logical block number for the MDB */
 
-#define kCalculatedExtentRefNum		( 4 )
-#define kCalculatedCatalogRefNum	(1*sizeof(SFCB) + 4)
-#define kCalculatedAllocationsRefNum	(2*sizeof(SFCB) + 4)
-#define kCalculatedAttributesRefNum	(3*sizeof(SFCB) + 4)
-#define kCalculatedRepairRefNum		(4*sizeof(SFCB) + 4)
+#define kCalculatedExtentRefNum			( 0 )
+#define kCalculatedCatalogRefNum		( 1*sizeof(SFCB) )
+#define kCalculatedAllocationsRefNum	( 2*sizeof(SFCB) )
+#define kCalculatedAttributesRefNum		( 3*sizeof(SFCB) )
+#define kCalculatedStartupRefNum		( 4*sizeof(SFCB) )
+#define kCalculatedRepairRefNum			( 5*sizeof(SFCB) )
 
 #define	Max_ABSiz	0x7FFFFE00		/* max allocation block size (multiple of 512 */
 #define	Blk_Size	512				/* size of a logical block */
+#define kHFSBlockSize 512			/* HFS block size */
 
 // only the lower 7 bits are considered to be invalid, all others are valid -djb
 #define	VAtrb_Msk	0x007F			/* volume attribute mask - invalid bits */
@@ -640,8 +648,10 @@ typedef struct SGlob {
 
 	Boolean			cleanUnmount;
 	Boolean			guiControl;
-	int			logLevel;
-	int			chkLevel;
+	int				logLevel;
+	int				chkLevel;
+	int             repairLevel;   // from 
+	BTScanState		scanState;
 
 	char			volumeName[256]; /* volume name in ASCII or UTF-8 */
 } SGlob, *SGlobPtr;
@@ -722,6 +732,7 @@ enum
 	repairLevelVolumeRecoverable,				//	Minor Volume corruption exists
 	repairLevelSomeDataLoss,					//	Overlapping extents, some data loss but no scavaging will get it back
 	repairLevelWillCauseDataLoss,				//	Missing leaf nodes, repair will lose nodes without scavaging (proceed at your own risk, check disk with other utils)
+	repairLevelCatalogBtreeRebuild,				//	Catalog Btree is damaged, repair may lose some data
 	repairLevelUnrepairable						//	DFA cannot repair volume
 };
 
@@ -980,6 +991,7 @@ extern	OSErr	AttrBTChk( SGlobPtr GPtr );		//	attributes btree check
 extern	OSErr	IVChk( SGlobPtr GPtr );
 
 extern	int	CheckForClean( SGlobPtr GPtr, Boolean markClean );
+extern  int	CheckIfJournaled(SGlobPtr GPtr);
 
 extern	OSErr	VInfoChk( SGlobPtr GPtr );
 
@@ -1017,6 +1029,12 @@ extern	OSErr	CmpBlock( void *block1P, void *block2P, UInt32 length ); /* same as
 	
 extern	OSErr	ChkExtRec ( SGlobPtr GPtr, const void *extents );
 
+
+/* -------------------------- From SRebuildCatalogBTree.c ------------------------- */
+
+extern	OSErr 	RebuildCatalogBTree( SGlobPtr theSGlobPtr );
+
+
 /* ------------------------------- From SExtents.c -------------------------------- */
 OSErr	ZeroFileBlocks( SVCB *vcb, SFCB *fcb, UInt32 startingSector, UInt32 numberOfSectors );
 
@@ -1028,14 +1046,16 @@ OSErr MapFileBlockC (
 	UInt64			*startSector,		// first 512-byte volume sector (NOT an allocation block)
 	UInt32			*availableBytes);	// number of contiguous bytes (up to numberOfBytes)
 
+OSErr DeallocateFile(SVCB *vcb, CatalogRecord * fileRec);
+
 OSErr ExtendFileC (
 	SVCB		*vcb,				// volume that file resides on
 	SFCB			*fcb,				// FCB of file to truncate
 	UInt32			sectorsToAdd,		// number of sectors to allocate
 	UInt32			flags,				// EFContig and/or EFAll
 	UInt32			*actualSectorsAdded); // number of bytes actually allocated
-
-OSErr DeallocateFile(SVCB *vcb, CatalogRecord * fileRec);
+	
+OSErr FlushExtentFile( SVCB *vcb );
 
 void ExtDataRecToExtents(
 	const HFSExtentRecord	oldExtents,
@@ -1053,7 +1073,7 @@ OSErr	CreateAttributesBTreeControlBlock( SGlobPtr GPtr );
 OSErr	CreateExtendedAllocationsFCB( SGlobPtr GPtr );
 
 
-OSErr	CacheWriteInPlace( SVCB *vcb, UInt32 fileRefNum,  HIOParam *iopb, UInt32 currentPosition,
+OSErr	CacheWriteInPlace( SVCB *vcb, UInt32 fileRefNum,  HIOParam *iopb, UInt64 currentPosition,
 	UInt32 maximumBytes, UInt32 *actualBytes );
 
 
@@ -1263,7 +1283,7 @@ extern int utf_encodestr(const u_int16_t *, size_t, u_int8_t *, size_t *);
  */
 extern int   HardLinkCheckBegin(SGlobPtr gp, void** cookie);
 extern void  HardLinkCheckEnd(void * cookie);
-extern void  CaptureHardLink(void * cookie, UInt32 linkID, UInt32 fileID);
+extern void  CaptureHardLink(void * cookie, UInt32 linkID);
 extern int   CheckHardLinks(void *cookie);
 
 /*
